/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 
 

package grph.io;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;

import com.carrotsearch.hppc.cursors.IntCursor;

import grph.Grph;
import grph.GrphAlgorithmCache;
import grph.algo.topology.ClassicalGraphs;
import grph.properties.Property;
import toools.UnitTests;
import toools.set.IntSet;

public class GrphBinaryWriter extends AbstractGraphWriter
{

	private boolean writeProperties = true;
	private boolean writeCache = true;

	@Override
	public void writeGraph(Grph graph, OutputStream os) throws IOException
	{
		assert graph != null;
		assert os != null;
		DataOutputStream dos = new DataOutputStream(os);
		dos.writeUTF(graph.getClass().getName());
		dos.writeBoolean(writeProperties);
		dos.writeBoolean(writeCache);
		writeTopology(graph, dos);

		if (writeProperties)
		{
			writeProperties(graph, dos);
		}

		if (writeCache)
		{
			writeCachedValues(graph, new ObjectOutputStream(dos));
		}
	}

	private void writeCachedValues(Grph graph, ObjectOutputStream dos) throws IOException
	{
		Collection<GrphAlgorithmCache> caches = graph.listCachingGraphAlgorithms();

		for (GrphAlgorithmCache cache : caches)
		{
			Object cachedValue = cache.getCachedValue();

			if (cachedValue != null && cachedValue instanceof Serializable)
			{
				dos.writeUTF(cache.getCachedAlgorithm().getClass().getName());
				dos.writeObject(cachedValue);
			}
		}

		dos.writeUTF("");
		dos.flush();
	}

	private void writeProperties(Grph g, DataOutputStream dos) throws IOException
	{
		Collection<Property> properties = g.getProperties();
		Iterator<Property> i = properties.iterator();

		while (i.hasNext())
		{
			if (i.next().getName() == null)
			{
				i.remove();
			}
		}

		dos.writeInt(properties.size());

		for (Property p : properties)
		{
			if (p.getName() != null)
			{
				dos.writeUTF(p.getName());
				p.toGrphBinary(dos);
			}
		}
	}

	private void writeTopology(Grph g, DataOutputStream dos) throws IOException
	{
		int numberOfVertices = g.getVertices().size();
		dos.writeInt(numberOfVertices);

		if (numberOfVertices > 0)
		{
			int greatestVertexID = g.getVertices().getGreatest();
			dos.writeInt(greatestVertexID);
			writeVertices(dos, g.getIsolatedVertices(), g);

			int numberOfEdges = g.getEdges().size();
			dos.writeInt(numberOfEdges);

			if (numberOfEdges > 0)
			{
				int greatestEdgeID = g.getEdges().getGreatest();
				dos.writeInt(greatestEdgeID);
				dos.writeInt(g.getNumberOfUndirectedSimpleEdges());

				for (int e : g.getEdges().toIntArray())
				{
					if (g.isUndirectedSimpleEdge(e))
					{
						writeEdge(dos, e, g);
						int a = g.getOneVertex(e);
						writeVertex(dos, a, g);
						writeVertex(dos, g.getTheOtherVertex(e, a), g);
					}
				}

				dos.writeInt(g.getNumberOfDirectedSimpleEdges());

				for (int e : g.getEdges().toIntArray())
				{
					if (g.isDirectedSimpleEdge(e))
					{
						writeEdge(dos, e, g);
						writeVertex(dos, g.getDirectedSimpleEdgeTail(e), g);
						writeVertex(dos, g.getDirectedSimpleEdgeHead(e), g);
					}
				}

				dos.writeInt(g.getNumberOfUndirectedHyperEdges());

				for (int e : g.getEdges().toIntArray())
				{
					if (g.isUndirectedHyperEdge(e))
					{
						writeEdge(dos, e, g);
						writeVertices(dos, g.getUndirectedHyperEdgeVertices(e), g);
					}
				}

				dos.writeInt(g.getNumberOfDirectedHyperEdges());

				for (int e : g.getEdges().toIntArray())
				{
					if (g.isDirectedHyperEdge(e))
					{
						writeEdge(dos, e, g);
						writeVertices(dos, g.getDirectedHyperEdgeTail(e), g);
						writeVertices(dos, g.getDirectedHyperEdgeHead(e), g);
					}
				}
			}
		}
	}

	public static void writeVertices(DataOutputStream dos, IntSet set, Grph g) throws IOException
	{
		dos.writeInt(set.size());

		for (IntCursor v : set)
		{
			writeVertex(dos, v.value, g);
		}
	}

	public static void writeEdge(DataOutputStream dos, int e, Grph g) throws IOException
	{
		writeInteger(e, dos, g.getEdges().getGreatest());
	}

	public static void writeVertex(DataOutputStream dos, int v, Grph g) throws IOException
	{
		writeInteger(v, dos, g.getVertices().getGreatest());
	}

	public static void writeInteger(int n, DataOutputStream dos, int greatestItem) throws IOException
	{
		if (greatestItem < 256)
		{
			dos.writeByte(n);
		}
		else if (greatestItem < 65536)
		{
			dos.writeChar(n);
		}
		else
		{
			dos.writeInt(n);
		}
	}

	private static void tdestBinaryEncoding() throws ParseException, GraphBuildException
	{
		Grph g = ClassicalGraphs.completeBipartiteGraph(10, 10);
		byte[] bytes = g.toGrphBinary();
		Grph h = Grph.fromGrphBinary(bytes);
		UnitTests.ensureEqual(h, g);
	}
}
