/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 
 package toools.collection.bigstuff.large;

import it.unimi.dsi.fastutil.longs.Long2ObjectMap;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;
import toools.UnitTests;
import toools.collections.LongIterator;
import toools.collections.ObjectIterator;

/**
 * Using % at first level prevents the close numbers to be all gathered in the
 * same submap
 * 
 * @author lhogie
 *
 */

public class LargeLongObjectOpenHashMap<T> extends AbstractLargeLongObjectMap<T>
{
	private Long2ObjectMap<Long2ObjectMap<T>> m = new Long2ObjectOpenHashMap<>();
	private T defaultValue;

	public LargeLongObjectOpenHashMap(long capacity)
	{
		super(capacity);
	}

	@Override
	public boolean containsKey(long key)
	{
		int slotNumber = getSlotNumber(key);
		Long2ObjectMap<T> slot = m.get(slotNumber);
		return slot != null && slot.containsKey(key);
	}

	@Override
	public T get(long key)
	{
		int slotNumber = getSlotNumber(key);
		Long2ObjectMap<T> slot = m.get(slotNumber);

		if (slot == null)
		{
			return defaultValue;
		}
		else
		{
			if ( ! slot.containsKey(key))
			{
				return defaultValue;
			}

			return slot.get(key);
		}
	}

	@Override
	public void put(long key, T v)
	{
		int slotNumber = getSlotNumber(key);
		Long2ObjectMap<T> slot = m.get(slotNumber);

		if (slot == null)
		{
			m.put(slotNumber, slot = new Long2ObjectOpenHashMap<T>(slotCapacity));
		}

		slot.put(key, v);
	}

	@Override
	public void remove(long key)
	{
		int slotNumber = getSlotNumber(key);
		Long2ObjectMap<T> slot = m.get(slotNumber);

		if (slot != null)
		{
			if (slot.containsKey(key))
			{
				slot.remove(key);

				if (slot.isEmpty())
				{
					m.remove(slotNumber);
				}
			}
		}
	}

	public long countKeys()
	{
		long size = 0;

		for (Long2ObjectMap<T> subMap : m.values())
		{
			size += subMap.size();
		}

		return size;
	}

	@Override
	public void clear()
	{
		m.clear();
	}

	@Override
	public long size()
	{
		return countKeys();
	}

	@Override
	public LongIterator iterator()
	{
		final it.unimi.dsi.fastutil.objects.ObjectIterator<Long2ObjectMap<T>> submapIterator = m
				.values().iterator();

		return new LongIterator()
		{
			Long2ObjectMap<T> currentSubmap = submapIterator.next();
			it.unimi.dsi.fastutil.longs.LongIterator currentSubmapKeyIterator = currentSubmap
					.keySet().iterator();

			@Override
			public long next()
			{
				assert currentSubmapKeyIterator.hasNext();
				long v = currentSubmapKeyIterator.next();

				// if we have reached the end of the submap
				if ( ! currentSubmapKeyIterator.hasNext())
				{
					// and there are other submaps to iterate over
					if (submapIterator.hasNext())
					{
						currentSubmap = submapIterator.next();
						currentSubmapKeyIterator = currentSubmap.keySet().iterator();
					}
					else
					{
						currentSubmap = null;
						currentSubmapKeyIterator = null;
					}
				}

				return v;
			}

			@Override
			public boolean hasNext()
			{
				return currentSubmapKeyIterator != null
						&& currentSubmapKeyIterator.hasNext();
			}
		};
	}

	@Override
	public ObjectIterator<T> values()
	{
		final it.unimi.dsi.fastutil.objects.ObjectIterator<Long2ObjectMap<T>> submapIterator = m
				.values().iterator();

		return new ObjectIterator<T>()
		{
			Long2ObjectMap<T> currentSubmap = submapIterator.next();
			it.unimi.dsi.fastutil.objects.ObjectIterator<T> currentSubmapKeyIterator = currentSubmap
					.values().iterator();

			@Override
			public T next()
			{
				T v = currentSubmapKeyIterator.next();

				if ( ! currentSubmapKeyIterator.hasNext())
				{
					if (submapIterator.hasNext())
					{
						currentSubmap = submapIterator.next();
						currentSubmapKeyIterator = currentSubmap.values().iterator();
					}
					else
					{
						currentSubmap = null;
						currentSubmapKeyIterator = null;
					}
				}

				return v;
			}

			@Override
			public boolean hasNext()
			{
				return currentSubmapKeyIterator != null
						&& currentSubmapKeyIterator.hasNext();
			}
		};
	}

	public static void main(String[] args)
	{
		LargeLongObjectOpenHashMap a = new LargeLongObjectOpenHashMap(0);
		System.out.println("size= " + a.size());
		a.put(2, 3);
		a.put(0, 1);

		LongIterator i = a.iterator();

		while (i.hasNext())
		{
			long k = i.next();
			System.out.println(k + " => " + a.get(k));
		}
	}

	private static void test()
	{
		LargeLongObjectOpenHashMap a = new LargeLongObjectOpenHashMap(0);
		a.put(2, 3);
		a.put(0, 1);
		UnitTests.ensureEqual(a.size(), 2);
		UnitTests.ensure(a.containsKey(2));
		UnitTests.ensure(a.containsKey(3));
	}
}
