/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.collection.bigstuff.longset;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import com.carrotsearch.hppc.LongOpenHashSet;

import it.unimi.dsi.fastutil.longs.LongArrayList;
import toools.BadIdea;
import toools.Clazz;
import toools.collections.LongIterator;
import toools.util.Conversion;

public abstract class LongSet implements Iterable<LongCursor>, Cloneable, Externalizable
{
	public static boolean allowCaching = true;
	private static List<LongSetListener> listeners;

	@Override
	public void writeExternal(ObjectOutput out) throws IOException
	{
		out.writeLong(size());

		for (LongCursor c : this)
		{
			out.writeLong(c.value);
		}
	}

	@Override
	public void readExternal(ObjectInput in) throws IOException
	{
		clear();
		long sz = in.readLong();

		for (long i = 0; i < sz; ++i)
		{
			add((int) in.readLong());
		}

	}

	public void addListener(LongSetListener l)
	{
		if (listeners == null)
		{
			listeners = new ArrayList<LongSetListener>();
		}

		listeners.add(l);
	}

	public void removeListener(LongSetListener l)
	{
		listeners.remove(l);

		if (listeners.isEmpty())
		{
			listeners = null;
		}
	}

	protected void fireAdd(long e)
	{
		if (listeners != null)
		{
			for (LongSetListener l : listeners)
			{
				l.elementAdded(this, e);
			}
		}
	}

	protected void fireRemove(long e)
	{
		if (listeners != null)
		{
			for (LongSetListener l : listeners)
			{
				l.elementRemoved(this, e);
			}
		}
	}

	public abstract boolean contains(long id);

	public abstract long size();

	public abstract boolean isEmpty();

	public abstract long pickRandomElement(Random prng);

	public long pickRandomElement(Random prng, boolean remove)
	{
		long e = pickRandomElement(prng);

		if (remove)
		{
			remove(e);
		}

		return e;
	}

	public LongIterator iteratorPrimitive()
	{
		return new LongIterator()
		{
			final Iterator<LongCursor> i = iterator();

			@Override
			public long next()
			{
				return i.next().value;
			}

			@Override
			public boolean hasNext()
			{
				return i.hasNext();
			}
		};
	}

	public abstract Class<?> getImplementationClass();

	public LongSet pickRandomSubset(Random prng, boolean remove)
	{
		return pickRandomSubset(prng, prng.nextLong() % size(), remove);
	}

	public LongSet pickRandomSubset(Random prng, long numberOfElements, boolean remove)
	{
		if (numberOfElements > size())
			throw new IllegalArgumentException("cannot pick that many elements");

		LongSet r = null;

		if (numberOfElements == size())
		{
			r = clone();

			if (remove)
			{
				clear();
			}
		}
		else
		{
			r = new LongHashSet();

			while (numberOfElements > r.size())
			{
				r.add(pickRandomElement(prng, r, remove));
			}
		}

		return r;
	}

	public long pickRandomElement(Random prng, long not, boolean remove)
	{
		while (true)
		{
			long e = pickRandomElement(prng);

			if (e != not)
			{
				if (remove)
				{
					remove(e);
				}

				return e;
			}
		}
	}

	public long pickRandomElement(Random prng, LongSet excludedElements, boolean remove)
	{
		while (true)
		{
			long e = pickRandomElement(prng);

			if ( ! excludedElements.contains(e))
			{
				if (remove)
				{
					remove(e);
				}

				return e;
			}
		}
	}

	public boolean isContiguous()
	{
		return getDensity() == 1;
	}

	public abstract long getGreatest();

	public final double getDensity()
	{
		long sizeMinusOne = size() - 1;
		long greatest = getGreatest();

		if (greatest == sizeMinusOne)
		{
			return 1;
		}
		else
		{
			return sizeMinusOne / (double) greatest;
		}
	}

	/**
	 * Modificators
	 */
	public abstract void add(long id);

	public abstract void remove(long id);

	public void remove(int... id)
	{
		for (long e : id)
		{
			remove(e);
		}
	}

	public abstract void clear();

	public void addAll(LongSet s)
	{
		for (LongCursor c : s)
		{
			add(c.value);
		}
	}

	public void addAll(long... ids)
	{
		for (long id : ids)
		{
			add(id);
		}
	}

	public void addAll(int... ids)
	{
		for (long id : ids)
		{
			add(id);
		}
	}

	public void removeAll(LongSet s)
	{
		for (LongCursor c : s)
		{
			remove(c.value);
		}
	}

	public void retainAll(LongSet s)
	{
		for (LongCursor c : this)
		{
			if ( ! s.contains(c.value))
			{
				remove(c.value);
			}
		}
	}

	@Override
	public boolean equals(Object o)
	{
		LongSet otherSet = (LongSet) o;

		if (otherSet.size() == size())
		{
			return contains(otherSet);
		}
		else
		{
			return false;
		}
	}

	public boolean contains(LongSet otherSet)
	{
		// a set can't contain a set larger than itself
		if (otherSet.size() > size())
		{
			return false;
		}
		else
		{
			for (LongCursor c : otherSet)
			{
				if ( ! contains(c.value))
				{
					return false;
				}
			}

			return true;
		}
	}

	@Override
	public String toString()
	{
		return toString(new ElementPrinter()
		{

			@Override
			public String toString(long e)
			{
				return String.valueOf(e);
			}
		});
	}

	public String toString(ElementPrinter p)
	{
		StringBuilder b = new StringBuilder();
		Iterator<LongCursor> i = iterator();
		b.append('{');

		while (i.hasNext())
		{
			b.append(p.toString(i.next().value));

			if (i.hasNext())
			{
				b.append(", ");
			}
		}

		b.append('}');
		return b.toString();
	}

	public String toString_numbers_only()
	{
		StringBuilder b = new StringBuilder();
		Iterator<LongCursor> i = iterator();

		while (i.hasNext())
		{
			b.append(i.next().value);

			if (i.hasNext())
			{
				b.append(' ');
			}
		}

		return b.toString();
	}

	public LongArrayList toLongArrayList()
	{
		LongArrayList l = new LongArrayList((int) size());

		for (LongCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	public LongOpenHashSet toLongOpenHashSet()
	{
		LongOpenHashSet l = new LongOpenHashSet();

		for (LongCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	public long[] toLongArray()
	{
		long[] array = new long[Conversion.long2int(size())];
		int i = 0;

		for (LongCursor c : this)
		{
			array[i++] = c.value;
		}

		assert i == size() : "i=" + i + ", size=" + size() + "  "
				+ getImplementationClass() + "  " + this;
		return array;
	}

	@Override
	public LongSet clone()
	{
		return clone(getClass());
	}

	public LongSet clone(Class<? extends LongSet> c)
	{
		LongSet s = Clazz.makeInstance(c);
		s.addAll(this);
		return s;
	}

	public boolean equals(long... ids)
	{
		LongSet s = new LongHashSet();
		s.addAll(ids);
		return equals(s);
	}

	@Override
	public final int hashCode()
	{
		throw new BadIdea("you better not call this because it uses a lot of time");
	}

	public void printTo(PrintWriter os)
	{
		Iterator<LongCursor> i = iterator();

		while (i.hasNext())
		{
			os.append(String.valueOf(i.next().value));

			if (i.hasNext())
			{
				os.append(' ');
			}
		}
	}

	public LongSet filter(LongSetFilter f)
	{
		LongSet r = new DefaultLongSet();

		for (long e : toLongArray())
		{
			if (f.accept(e))
			{
				r.add(e);
			}
		}

		return r;
	}
}
