/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.collection.bigstuff.longset;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;

import it.unimi.dsi.fastutil.longs.LongArrayList;
import toools.Clazz;
import toools.UnitTests;
import toools.math.MathsUtilities;

public class LongSets
{
	public static final LongSet emptySet = EmptySet.instance;

	public static <S extends LongSet> S union(Class<S> target, LongSet... sets)
	{
		S c = Clazz.makeInstance(target);

		for (LongSet set : sets)
		{
			c.addAll(set);
		}

		return c;
	}

	public static LongSet union(LongSet... sets)
	{
		return union(DefaultLongSet.class, sets);
	}

	public static <S extends LongSet> S createRandomSet(Class<S> target, Random r, long n,
			long upperbound)
	{
		if (upperbound < n)
			throw new IllegalArgumentException(
					"not enough elements available to create such a large set");

		S s = Clazz.makeInstance(target);

		while (s.size() < n)
		{
			s.add(r.nextLong() % upperbound);
		}

		return s;
	}

	public static LongSet singleton(long n)
	{
		LongSet s = new LongSingletonSet(n);
		return s;
	}

	public static LongArrayList getRandomSubset(LongArrayList elements,
			long numberOfElementsRequested, Random random)
	{
		if (numberOfElementsRequested > elements.size())
			throw new IllegalArgumentException("you ask for too much");

		return getElementAtIndexes(elements,
				getRandomIndexes(elements.size(), numberOfElementsRequested, random));
	}

	private static LongArrayList getRandomIndexes(long numberOfElements,
			long numberOfIndicesRequested, Random random)
	{
		LongArrayList indexes = new LongArrayList();

		while (indexes.size() < numberOfIndicesRequested)
		{
			indexes.add(
					(int) MathsUtilities.pickRandomBetween(0, numberOfElements, random));
		}

		return indexes;
	}

	private static LongArrayList getElementAtIndexes(LongArrayList elements,
			LongArrayList indexes)
	{
		LongArrayList res = new LongArrayList();

		for (long i = 0; i < indexes.size(); ++i)
		{
			res.add(elements.get((int) i));
		}

		return res;
	}

	public static long pickRandomElement(LongArrayList l, Random random)
	{
		return l.get(random.nextInt(l.size()));
	}


	public static long findMax(LongSet s)
	{
		assert ! s.isEmpty();
		Iterator<LongCursor> i = s.iterator();
		long max = i.next().value;

		while (i.hasNext())
		{
			max = i.next().value;
		}

		return max;
	}

	public static LongSet from(int... array)
	{
		return from(DefaultLongSet.class, array);
	}

	public static LongSet from(long... array)
	{
		return from(DefaultLongSet.class, array);
	}

	public static <S extends LongSet> S from(Class<S> c, int... array)
	{
		S s = Clazz.makeInstance(c);
		s.addAll(array);
		return s;
	}

	public static <S extends LongSet> S from(Class<S> c, long... array)
	{
		S s = Clazz.makeInstance(c);
		s.addAll(array);
		return s;
	}

	public static <S extends LongSet> S intersection(Class<S> target, LongSet a,
			LongSet b)
	{
		LongSet smallestSet = a.size() < b.size() ? a : b;
		LongSet greatestSet = smallestSet == a ? b : a;

		S r = Clazz.makeInstance(target);

		for (LongCursor c : smallestSet)
		{
			if (greatestSet.contains(c.value))
			{
				r.add(c.value);
			}
		}

		return r;
	}

	public static LongSet intersection(LongSet a, LongSet b)
	{
		return intersection(DefaultLongSet.class, a, b);
	}

	/**
	 * @param a
	 * @param b
	 * @return an IntSet that contains all elements of a that do not appear in
	 *         set b
	 */
	public static <S extends LongSet> S difference(Class<S> target, LongSet a, LongSet b)
	{
		S diff = Clazz.makeInstance(target);

		for (LongCursor i : a)
			if ( ! b.contains(i.value))
				diff.add(i.value);

		return diff;
	}

	public static LongSet difference(LongSet a, LongSet b)
	{
		return difference(DefaultLongSet.class, a, b);
	}

	/**
	 * Sorts a list of sets by ascending cardinality.
	 * 
	 * @param tab
	 *            a list of IntSets
	 * @param s
	 *            start index
	 * @param e
	 *            end index
	 */
	public static void quickSortSet(ArrayList<LongSet> tab, long s, long e)
	{
		if (s < e)
		{
			long m = partitionSet(tab, s, e);
			quickSortSet(tab, s, m);
			quickSortSet(tab, m + 1, e);
		}
	}

	private static long partitionSet(ArrayList<LongSet> tab, long s, long e)
	{
		long i = s - 1;
		long j = e + 1;
		long pivot = tab.get((int) s).size();

		LongSet temp;
		while (true)
		{
			do
				j--;
			while (tab.get((int) j).size() > pivot);

			do
				i++;
			while (tab.get((int) i).size() < pivot);

			if (i < j)
			{
				temp = new DefaultLongSet();
				temp.addAll(tab.get((int) i));
				tab.set((int) i, tab.get((int) j));
				tab.set((int) j, temp);
			}
			else
				return j;
		}
	}

	/**
	 * Parse the given strings, expecting that is describes a set of ints.
	 * Non-numerical characters are ignored. This enables the parsing of: [4, 5,
	 * 6] [4 5 6] 4, 5, 6 4 5 6 {4 5 6}
	 * 
	 * @param a
	 *            string describing a set of positive integers.
	 * @return the set of positive integers
	 */
	public static LongSet from(String s)
	{
		LongSet r = new DefaultLongSet();

		for (String t : s.replaceAll("[^0-9]", " ").trim().split(" +"))
		{
			r.add(Integer.parseInt(t));
		}

		return r;
	}

	public static LongArrayList toArrayList(String s)
	{
		LongArrayList r = new LongArrayList();

		for (String t : s.replaceAll("[^0-9]", " ").trim().split(" +"))
		{
			r.add(Long.parseLong(t));
		}

		return r;
	}

	private static void testParse()
	{
		from("[5 7 89]");
	}

	public static void main(String[] args)
	{
		System.out.println(from("[5 7 89]"));
	}

	public static LongSet from(Collection<Integer> c)
	{
		return from(DefaultLongSet.class, c);
	}

	public static <S extends LongSet> S from(Class<S> target, Collection<Integer> c)
	{
		S r = Clazz.makeInstance(target);

		for (Integer i : c)
		{
			r.add(i);
		}

		return r;
	}

	private static void testDensity()
	{
		LongSet s = new DefaultLongSet();
		s.addAll(0, 1, 2, 3);
		UnitTests.ensure(s.getDensity() == 1);
	}


	private static void test()
	{
		Collection<LongSet> sets = new ArrayList();
		sets.add(new LongHashSet());

		for (LongSet s : sets)
		{
			// Grph.logger.log("testing " + s.getClass());
			s.addAll(1, 2, 3, 4);
			UnitTests.ensureEqual(s.size(), 4);
			UnitTests.ensureEqual(s.getGreatest(), 4);
			UnitTests.ensure(s.getDensity() < 1);
			s.add(0);
			UnitTests.ensure(s.getDensity() == 1);

			s.add(6);
			UnitTests.ensureEqual(s.getGreatest(), 6);

			s.add(0);
			s.add(1);
			s.add(5);
			UnitTests.ensure(s.getDensity() == 1);
		}

	}

}
