/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.set;

import java.util.BitSet;
import java.util.Iterator;
import java.util.Random;

import toools.UnitTests;

import com.carrotsearch.hppc.cursors.IntCursor;

public class BitVectorSet extends IntSet
{
	private BitSet bitset = new BitSet();
	private long len = 0;
	private final int offset;

	public BitVectorSet(int offset)
	{
		this.offset = offset;
	}

	@Override
	public boolean contains(int id)
	{
		assert id >= 0;
		return bitset.get(id - offset);

		// if (id / 64 >= bitset.bits.length)
		// return false;
		//
		// long cell = bitset.bits[id / 64];
		//
		// if (cell == 0)
		// return false;
		//
		// return ((cell >> (id % 64)) & 1) > 0;
	}

	@Override
	public Iterator<IntCursor> iterator()
	{

		return new Iterator<IntCursor>() {
			int indexOfNextSetBit = bitset.nextSetBit(0);
			final IntCursor hppcCursor = new IntCursor();

			@Override
			public boolean hasNext()
			{
				return indexOfNextSetBit >= 0;
			}

			@Override
			public IntCursor next()
			{
				hppcCursor.index = hppcCursor.value = indexOfNextSetBit + offset;
				indexOfNextSetBit = bitset.nextSetBit(indexOfNextSetBit + 1);
				return hppcCursor;
			}

			@Override
			public void remove()
			{
				throw new IllegalStateException();
			}
		};
	}

	@Override
	public int size()
	{
		return (int) len;
	}

	@Override
	public boolean isEmpty()
	{
		return len == 0;
	}

	@Override
	public void add(int id)
	{
		assert id >= 0;
		id -= offset;

		// if the element was not already in
		if (!bitset.get(id))
		{
			++len;
		}

		bitset.set(id);
	}

	@Override
	public void addAll(IntSet s)
	{
		BitVectorSet bvs = getUnderlyingBitSet_if_any(s);

		if (bvs == null || bvs.offset != offset)
		{
			super.addAll(s);
		}
		else
		{
			bitset.or(bvs.bitset);
			len = bitset.cardinality();
		}
	}

	/**
	 * Returns the bitSet underlying in the set implementation, if any.
	 * 
	 * @param s
	 * @return
	 */
	private static BitVectorSet getUnderlyingBitSet_if_any(IntSet s)
	{
		if (s.getClass() == BitVectorSet.class)
		{
			return (BitVectorSet) s;
		}
		else if (s instanceof SelfAdaptiveIntSet)
		{
			return getUnderlyingBitSet_if_any(((SelfAdaptiveIntSet) s).underlyingSet);
		}
		else
		{
			return null;
		}
	}

	@Override
	public void retainAll(IntSet s)
	{
		BitVectorSet bvs = getUnderlyingBitSet_if_any(s);

		if (bvs == null || bvs.offset != offset)
		{
			super.retainAll(s);
		}
		else
		{
			bitset.and(bvs.bitset);
			len = bitset.cardinality();
		}
	}

	@Override
	public void removeAll(IntSet s)
	{
		BitVectorSet bvs = getUnderlyingBitSet_if_any(s);

		if (bvs == null || bvs.offset != offset)
		{
			super.removeAll(s);
		}
		else
		{
			BitSet sb = (BitSet) bvs.bitset.clone();
			sb.flip(0, bitset.length());
			bitset.and(sb);
			len = bitset.cardinality();
		}
	}

	@Override
	public void remove(int id)
	{
		assert id >= 0;
		id -= offset;

		if (!bitset.get(id))
			throw new IllegalArgumentException("this set doesn't contains element " + id);

		bitset.clear(id);
		--len;
	}

	/*
	 * @Override public double getDensity() { double c = bs.cardinality();
	 * double l = bs.capacity(); return c / l; }
	 */
	@Override
	public void clear()
	{
		bitset.clear();
		len = 0;
	}

	@Override
	public int pickRandomElement(Random prng)
	{
		while (true)
		{
			int i = prng.nextInt((int) bitset.size());

			if (bitset.get(i))
			{
				return i + offset;
			}
		}
	}

	@Override
	public int getGreatest()
	{
		return (int) bitset.length() - 1 + offset;
	}

	public String toBitString()
	{
		StringBuilder b = new StringBuilder();

		for (int i = 0; i < bitset.length(); ++i)
		{
			if (i % 4 == 0)
			{
				b.append(' ');
			}

			if (i % 8 == 0)
			{
				b.append(' ');
			}

			b.append(bitset.get(i) ? '1' : '0');
		}

		return b.toString();
	}

	private static void testHashcode()
	{
		IntSet s = new BitVectorSet(0);
		s.addAll(1, 2, 3);

		IntSet s2 = new IntHashSet();
		s2.addAll(1, 3, 2);
		UnitTests.ensure(s.hashCode() == s2.hashCode());

	}
	

	@Override
	public Class<?> getImplementationClass()
	{
		return BitSet.class;
	}

}
