/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.set;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import com.carrotsearch.hppc.IntArrayList;
import com.carrotsearch.hppc.IntOpenHashSet;
import com.carrotsearch.hppc.cursors.IntCursor;

import toools.Clazz;
import toools.math.IntIterator;

public abstract class IntSet implements Iterable<IntCursor>, Cloneable, Externalizable
{
	public static boolean allowCaching = true;
	private static List<IntSetListener> listeners;

	@Override
	public void writeExternal(ObjectOutput out) throws IOException
	{
		out.writeLong(size());

		for (IntCursor c : this)
		{
			out.writeLong(c.value);
		}
	}

	@Override
	public void readExternal(ObjectInput in) throws IOException
	{
		clear();
		long sz = in.readLong();

		for (long i = 0; i < sz; ++i)
		{
			add((int) in.readLong());
		}

	}

	public void addListener(IntSetListener l)
	{
		if (listeners == null)
		{
			listeners = new ArrayList<IntSetListener>();
		}

		listeners.add(l);
	}

	public void removeListener(IntSetListener l)
	{
		listeners.remove(l);

		if (listeners.isEmpty())
		{
			listeners = null;
		}
	}

	protected void fireAdd(int e)
	{
		if (listeners != null)
		{
			for (IntSetListener l : listeners)
			{
				l.elementAdded(this, e);
			}
		}
	}

	protected void fireRemove(int e)
	{
		if (listeners != null)
		{
			for (IntSetListener l : listeners)
			{
				l.elementRemoved(this, e);
			}
		}
	}

	public abstract boolean contains(int id);

	public abstract int size();

	public abstract boolean isEmpty();

	public abstract int pickRandomElement(Random prng);

	public int pickRandomElement(Random prng, boolean remove)
	{
		int e = pickRandomElement(prng);

		if (remove)
		{
			remove(e);
		}

		return e;
	}

	public IntIterator iteratorPrimitive()
	{
		return new IntIterator()
		{
			final Iterator<IntCursor> i = iterator();

			@Override
			public int next()
			{
				return i.next().value;
			}

			@Override
			public boolean hasNext()
			{
				return i.hasNext();
			}
		};
	}

	public abstract Class<?> getImplementationClass();

	public IntSet pickRandomSubset(Random prng, boolean remove)
	{
		return pickRandomSubset(prng, prng.nextInt(size()), remove);
	}

	public IntSet pickRandomSubset(Random prng, int numberOfElements, boolean remove)
	{
		if (numberOfElements > size())
			throw new IllegalArgumentException("cannot pick that many elements");

		IntSet r = null;

		if (numberOfElements == size())
		{
			r = clone();

			if (remove)
			{
				clear();
			}
		}
		else
		{
			r = new IntHashSet();

			while (numberOfElements > r.size())
			{
				r.add(pickRandomElement(prng, r, remove));
			}
		}

		return r;
	}

	public int pickRandomElement(Random prng, int not, boolean remove)
	{
		while (true)
		{
			int e = pickRandomElement(prng);

			if (e != not)
			{
				if (remove)
				{
					remove(e);
				}

				return e;
			}
		}
	}

	public int pickRandomElement(Random prng, IntSet excludedElements, boolean remove)
	{
		while (true)
		{
			int e = pickRandomElement(prng);

			if ( ! excludedElements.contains(e))
			{
				if (remove)
				{
					remove(e);
				}

				return e;
			}
		}
	}

	public boolean isContiguous()
	{
		return getDensity() == 1;
	}

	public abstract int getGreatest();

	public final double getDensity()
	{
		int sizeMinusOne = size() - 1;
		int greatest = getGreatest();

		if (greatest == sizeMinusOne)
		{
			return 1;
		}
		else
		{
			return sizeMinusOne / (double) greatest;
		}
	}

	/**
	 * Modificators
	 */
	public abstract void add(int id);

	public abstract void remove(int id);

	public void remove(int... id)
	{
		for (int e : id)
		{
			remove(e);
		}
	}

	public abstract void clear();

	public void addAll(IntSet s)
	{
		for (IntCursor c : s)
		{
			add(c.value);
		}
	}


	public void addAll(int... ids)
	{
		for (int id : ids)
		{
			add(id);
		}
	}

	public void removeAll(IntSet s)
	{
		for (IntCursor c : s)
		{
			remove(c.value);
		}
	}

	public void retainAll(IntSet s)
	{
		for (IntCursor c : this)
		{
			if ( ! s.contains(c.value))
			{
				remove(c.value);
			}
		}
	}

	@Override
	public boolean equals(Object o)
	{
		if (o == this)
			return true;

		if (o instanceof IntSet)
			return equals((IntSet) o);

		return false;
	}

	public boolean equals(IntSet otherSet)
	{
		return otherSet.size() == size() && contains(otherSet);
	}

	public boolean contains(IntSet otherSet)
	{
		// a set can't contain a set larger than itself
		if (otherSet.size() > size())
		{
			return false;
		}
		else
		{
			for (IntCursor c : otherSet)
			{
				if ( ! contains(c.value))
				{
					return false;
				}
			}

			return true;
		}
	}

	@Override
	public String toString()
	{
		return toString(new ElementPrinter()
		{

			@Override
			public String toString(int e)
			{
				return String.valueOf(e);
			}
		});
	}

	public String toString(ElementPrinter p)
	{
		StringBuilder b = new StringBuilder();
		Iterator<IntCursor> i = iterator();
		b.append('{');

		while (i.hasNext())
		{
			b.append(p.toString(i.next().value));

			if (i.hasNext())
			{
				b.append(", ");
			}
		}

		b.append('}');
		return b.toString();
	}

	public String toString_numbers_only()
	{
		StringBuilder b = new StringBuilder();
		Iterator<IntCursor> i = iterator();

		while (i.hasNext())
		{
			b.append(i.next().value);

			if (i.hasNext())
			{
				b.append(' ');
			}
		}

		return b.toString();
	}

	public IntArrayList toIntArrayList()
	{
		IntArrayList l = new IntArrayList();

		for (IntCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	public IntSequence toIntSequence()
	{
		IntSequence l = new IntSequence();

		for (IntCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	public IntOpenHashSet toIntOpenHashSet()
	{
		IntOpenHashSet l = new IntOpenHashSet();

		for (IntCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	/**
	 * Be careful. Use a lot of memory.
	 * 
	 * @return
	 */
	public ArrayList<Integer> toIntegerArrayList()
	{
		ArrayList<Integer> l = new ArrayList();

		for (IntCursor c : this)
		{
			l.add(c.value);
		}

		return l;
	}

	public int[] toIntArray()
	{
		int[] array = new int[size()];
		int i = 0;

		for (IntCursor c : this)
		{
			array[i++] = c.value;
		}

		assert i == size() : "i=" + i + ", size=" + size() + "  "
				+ getImplementationClass() + "  " + this;
		return array;
	}

	@Override
	public IntSet clone()
	{
		return clone(getClass());
	}

	public IntSet clone(Class<? extends IntSet> c)
	{
		IntSet s = Clazz.makeInstance(c);
		s.addAll(this);
		return s;
	}

	public boolean equals(int... ids)
	{
		IntSet s = new IntHashSet();
		s.addAll(ids);
		return equals(s);
	}

	@Override
	public final int hashCode()
	{
		// taken from Arrays.hashCode()
		int result = 1;
		for (IntCursor element : this)
		{
			int elementHash = (int) (element.value ^ (element.value >>> 32));
			result = 31 * result + elementHash;
		}

		return result;
	}

	public void writeTo(PrintStream os)
	{
		Iterator<IntCursor> i = iterator();

		while (i.hasNext())
		{
			os.append(String.valueOf(i.next().value));

			if (i.hasNext())
			{
				os.append(' ');
			}
		}
	}

	public IntSet filter(IntSetFilter f)
	{
		IntSet r = new DefaultIntSet();

		for (int e : toIntArray())
		{
			if (f.accept(e))
			{
				r.add(e);
			}
		}

		return r;
	}

	public long[] toLongArray()
	{
		long[] array = new long[size()];
		int i = 0;

		for (IntCursor c : this)
		{
			array[i++] = c.value;
		}

		assert i == size() : "i=" + i + ", size=" + size() + "  "
				+ getImplementationClass() + "  " + this;
		return array;
	}

}
