/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.set;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Random;

import toools.Clazz;
import toools.UnitTests;
import toools.math.MathsUtilities;

import com.carrotsearch.hppc.IntArrayList;
import com.carrotsearch.hppc.cursors.IntCursor;

public class IntSets
{
	public static final IntSet emptySet = EmptySet.instance;

	public static <S extends IntSet> S union(Class<S> target, IntSet... sets)
	{
		S c = Clazz.makeInstance(target);

		for (IntSet set : sets)
		{
			c.addAll(set);
		}

		return c;
	}

	public static IntSet union(IntSet... sets)
	{
		return union(DefaultIntSet.class, sets);
	}

	public static <S extends IntSet> S createRandomSet(Class<S> target, Random r, int n, int upperbound)
	{
		if (upperbound < n)
			throw new IllegalArgumentException("not enough elements available to create such a large set");

		S s = Clazz.makeInstance(target);

		while (s.size() < n)
		{
			s.add(r.nextInt(upperbound));
		}

		return s;
	}

	public static IntSet singleton(int n)
	{
		IntSet s = new IntSingletonSet(n);
		return s;
	}

	public static IntArrayList getRandomSubset(IntArrayList elements, int numberOfElementsRequested, Random random)
	{
		if (numberOfElementsRequested > elements.size())
			throw new IllegalArgumentException("you ask for too much");

		return getElementAtIndexes(elements, getRandomIndexes(elements.size(), numberOfElementsRequested, random));
	}

	private static IntArrayList getRandomIndexes(int numberOfElements, int numberOfIndicesRequested, Random random)
	{
		IntArrayList indexes = new IntArrayList();

		while (indexes.size() < numberOfIndicesRequested)
		{
			indexes.add((int) MathsUtilities.pickRandomBetween(0, numberOfElements, random));
		}

		return indexes;
	}

	private static IntArrayList getElementAtIndexes(IntArrayList elements, IntArrayList indexes)
	{
		IntArrayList res = new IntArrayList();

		for (int i = 0; i < indexes.size(); ++i)
		{
			res.add(elements.get(i));
		}

		return res;
	}

	public static int pickRandomElement(IntArrayList l, Random random)
	{
		return l.get(random.nextInt(l.size()));
	}

	public static int findMax(IntSet s)
	{
		assert !s.isEmpty();
		Iterator<IntCursor> i = s.iterator();
		int max = i.next().value;

		while (i.hasNext())
		{
			max = i.next().value;
		}

		return max;
	}

	public static IntSet from(int... array)
	{
		return from(SelfAdaptiveIntSet.class, array);
	}

	public static <S extends IntSet> S from(Class<S> c, int... array)
	{
		S s = Clazz.makeInstance(c);
		s.addAll(array);
		return s;
	}

	public static <S extends IntSet> S intersection(Class<S> target, IntSet a, IntSet b)
	{
		IntSet smallestSet = a.size() < b.size() ? a : b;
		IntSet greatestSet = smallestSet == a ? b : a;

		S r = Clazz.makeInstance(target);

		for (IntCursor c : smallestSet)
		{
			if (greatestSet.contains(c.value))
			{
				r.add(c.value);
			}
		}

		return r;
	}

	public static IntSet intersection(IntSet a, IntSet b)
	{
		return intersection(SelfAdaptiveIntSet.class, a, b);
	}

	/**
	 * @param a
	 * @param b
	 * @return an IntSet that contains all elements of a that do not appear in
	 *         set b
	 */
	public static <S extends IntSet> S difference(Class<S> target, IntSet a, IntSet b)
	{
		S diff = Clazz.makeInstance(target);

		for (IntCursor i : a)
			if (!b.contains(i.value))
				diff.add(i.value);

		return diff;
	}

	public static IntSet difference(IntSet a, IntSet b)
	{
		return difference(DefaultIntSet.class, a, b);
	}

	/**
	 * Sorts a list of sets by ascending cardinality.
	 * 
	 * @param tab
	 *            a list of IntSets
	 * @param s
	 *            start index
	 * @param e
	 *            end index
	 */
	public static void quickSortSet(ArrayList<IntSet> tab, int s, int e)
	{
		if (s < e)
		{
			int m = partitionSet(tab, s, e);
			quickSortSet(tab, s, m);
			quickSortSet(tab, m + 1, e);
		}
	}

	private static int partitionSet(ArrayList<IntSet> tab, int s, int e)
	{
		int i = s - 1;
		int j = e + 1;
		int pivot = tab.get(s).size();

		IntSet temp;
		while (true)
		{
			do
				j--;
			while (tab.get(j).size() > pivot);

			do
				i++;
			while (tab.get(i).size() < pivot);

			if (i < j)
			{
				temp = new DefaultIntSet(0);
				temp.addAll(tab.get(i));
				tab.set(i, tab.get(j));
				tab.set(j, temp);
			}
			else
				return j;
		}
	}

	/**
	 * Parse the given strings, expecting that is describes a set of ints.
	 * Non-numerical characters are ignored. This enables the parsing of: [4, 5,
	 * 6] [4 5 6] 4, 5, 6 4 5 6 {4 5 6}
	 * 
	 * @param a
	 *            string describing a set of positive integers.
	 * @return the set of positive integers
	 */
	public static IntSet from(String s)
	{
		IntSet r = new DefaultIntSet(0);

		for (String t : s.replaceAll("[^0-9]", " ").trim().split(" +"))
		{
			r.add(Integer.parseInt(t));
		}

		return r;
	}

	public static IntArrayList toArrayList(String s)
	{
		IntArrayList r = new IntArrayList();

		for (String t : s.replaceAll("[^0-9]", " ").trim().split(" +"))
		{
			r.add(Integer.parseInt(t));
		}

		return r;
	}

	private static void testParse()
	{
		from("[5 7 89]");
	}

	public static void main(String[] args)
	{
		System.out.println(from("[5 7 89]"));
	}

	public static IntSet from(Collection<Integer> c)
	{
		return from(SelfAdaptiveIntSet.class, c);
	}

	public static <S extends IntSet> S from(Class<S> target, Collection<Integer> c)
	{
		S r = Clazz.makeInstance(target);

		for (Integer i : c)
		{
			r.add(i);
		}

		return r;
	}

	private static void testDensity()
	{
		IntSet s = new DefaultIntSet(0);
		s.addAll(0, 1, 2, 3);
		UnitTests.ensure(s.getDensity() == 1);
	}

	private static void testIDSets()
	{
		IntSet s = new BitVectorSet(50);
		UnitTests.ensure(s.isEmpty());
		s.add(0);
		s.add(1);
		s.add(3);
		s.add(2);

		IntSet s2 = new IntHashSet();
		UnitTests.ensure(s2.isEmpty());

		s2.add(0);
		s2.add(1);
		s2.add(2);
		s2.add(3);
		UnitTests.ensure(s.equals(s2));

		UnitTests.ensure(s.contains(3));
		s.remove(3);
		UnitTests.ensure(!s.contains(3));
	}

	private static void test()
	{
		Collection<IntSet> sets = new ArrayList();
		sets.add(new IntHashSet());
		sets.add(new BitVectorSet(32));

		for (IntSet s : sets)
		{
			// Grph.logger.log("testing " + s.getClass());
			s.addAll(1, 2, 3, 4);
			UnitTests.ensureEqual(s.size(), 4);
			UnitTests.ensureEqual(s.getGreatest(), 4);
			UnitTests.ensure(s.getDensity() < 1);
			s.add(0);
			UnitTests.ensure(s.getDensity() == 1);

			s.add(6);
			UnitTests.ensureEqual(s.getGreatest(), 6);

			s.add(0);
			s.add(1);
			s.add(5);
			UnitTests.ensure(s.getDensity() == 1);
		}

	}

}
