/* (C) Copyright 2009-2013 CNRS (Centre National de la Recherche Scientifique).

Licensed to the CNRS under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The CNRS licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.

*/

/* Contributors:

Luc Hogie (CNRS, I3S laboratory, University of Nice-Sophia Antipolis) 
Aurelien Lancin (Coati research team, Inria)
Christian Glacet (LaBRi, Bordeaux)
David Coudert (Coati research team, Inria)
Fabien Crequis (Coati research team, Inria)
Grégory Morel (Coati research team, Inria)
Issam Tahiri (Coati research team, Inria)
Julien Fighiera (Aoste research team, Inria)
Laurent Viennot (Gang research-team, Inria)
Michel Syska (I3S, Université Cote D'Azur)
Nathann Cohen (LRI, Saclay) 
Julien Deantoin (I3S, Université Cote D'Azur, Saclay) 

*/
 


package toools.set;

import java.util.Iterator;
import java.util.Random;

import toools.Clazz;
import toools.thread.Threads;

import com.carrotsearch.hppc.cursors.IntCursor;

public class SelfAdaptiveIntSet extends IntSet
{
	// BitVectorSet needs to access this for optimization purpose
	IntSet underlyingSet;
	final private int offset;

	public SelfAdaptiveIntSet()
	{
		this(0);
	}

	
	public SelfAdaptiveIntSet(int offset)
	{
		this.offset = offset;
	}

	@Override
	public void add(int e)
	{
		if (underlyingSet == null)
		{
			if (e < 64)
			{
				underlyingSet = new BitVectorSet(offset);
			}
			else
			{
				underlyingSet = new IntHashSet();
			}
		}

		double r = (8d * size()) / Math.max(e, getGreatest());

		if (r < 0.9)
		{
			ensureHashset();
		}
		else if (r > 1.1)
		{
			ensureBitset();
		}

		underlyingSet.add(e);
		fireAdd(e);
	}

	@Override
	public void remove(int e)
	{
		if (underlyingSet == null)
		{
			throw new IllegalArgumentException("this set doesn't contains element " + e);
		}
		else
		{
			underlyingSet.remove(e);

			if (underlyingSet.isEmpty())
			{
				underlyingSet = null;
			}
			else
			{
				double r = (8d * size()) / getGreatest();

				if (r < 0.9)
				{
					ensureHashset();
				}
				else if (r > 1.1)
				{
					ensureBitset();
				}
			}

			fireRemove(e);
		}
	}

	@Override
	public boolean contains(int id)
	{
		return underlyingSet != null && underlyingSet.contains(id);
	}

	@Override
	public int size()
	{
		return underlyingSet == null ? 0 : underlyingSet.size();
	}

	@Override
	public boolean isEmpty()
	{
		return underlyingSet == null;
	}

	@Override
	public Iterator<IntCursor> iterator()
	{
		return underlyingSet == null ? IntSets.emptySet.iterator() : underlyingSet.iterator();
	}

	@Override
	public void clear()
	{
		if (underlyingSet != null)
		{
			underlyingSet = null;
		}
	}

	@Override
	public int pickRandomElement(Random prng)
	{
		return underlyingSet.pickRandomElement(prng);
	}

	private void ensureHashset()
	{
		if (underlyingSet.getClass() != IntHashSet.class)
		{
			IntSet newset = new IntHashSet();
			newset.addAll(underlyingSet);
			underlyingSet = newset;
		}
	}

	private void ensureBitset()
	{
		if (underlyingSet.getClass() != BitVectorSet.class)
		{
			IntSet newset = new BitVectorSet(offset);
			newset.addAll(underlyingSet);
			underlyingSet = newset;
		}
	}

	@Override
	public int getGreatest()
	{
		assert underlyingSet != null : "can't take the greatest element of an empty set";
		
		return underlyingSet.getGreatest();
	}

	public static void main(String[] args)
	{
		Clazz.makeInstance(SelfAdaptiveIntSet.class);

		DefaultIntSet set = new DefaultIntSet(0);
		set.add(2);
		Random r = new Random();

		System.out.println("****************");

		for (int i = 0; i < 10000000; ++i)
		{
			int b = r.nextInt(100);
			set.add(b);
			System.out.println("adding " + b);

			System.out.print(set);

			System.out.println(set.getImplementationClass());
			System.out.println(" size=" + set.size());
			System.out.println(" density=" + set.getDensity());
			Threads.sleepMs(100);
		}
	}

	public Class<? extends IntSet> getImplementationClass()
	{
		return underlyingSet == null ? null : underlyingSet.getClass();
	}
}
